package com.inteligr8.git;

import java.io.Serializable;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

// TODO need to implement the actual expiration schedule
public class LRUExpiringHashMap<K extends Serializable, V> implements ListeningMap<K, V> {

	private final Logger logger = LoggerFactory.getLogger(LRUExpiringHashMap.class);
	
	private final Object sync = new Object();
	private final LinkedHashMap<K, ExpiringValue> map;
	private final List<MapListener<K, V>> listeners = new LinkedList<MapListener<K, V>>();
	private final long expirationTimeMillis;
	
	public LRUExpiringHashMap(int expirationTimeInMinutes) {
		this.map = new LinkedHashMap<>();
		this.expirationTimeMillis = expirationTimeInMinutes * 60L + 1000L;
	}

	public LRUExpiringHashMap(int expirationTimeInMinutes, int initialCapacity) {
		this.map = new LinkedHashMap<>(initialCapacity);
		this.expirationTimeMillis = expirationTimeInMinutes * 60L + 1000L;
	}
	
	@Override
	public void addListener(MapListener<K, V> listener) {
    	if (this.logger.isDebugEnabled())
    		this.logger.debug("adding listener: " + listener.getClass());
		this.listeners.add(listener);
	}
	
	@Override
	public boolean containsKey(Object key) {
		return this.map.containsKey(key);
	}
	
	@Override
	public boolean containsValue(Object value) {
		return this.map.containsValue(value);
	}
	
	@Override
	public Set<Entry<K, V>> entrySet() {
		throw new UnsupportedOperationException();
	}
	
	@Override
	public Set<K> keySet() {
		throw new UnsupportedOperationException();
	}
	
	@Override
	public Collection<V> values() {
		throw new UnsupportedOperationException();
	}
	
	@Override
	public boolean isEmpty() {
		return this.map.isEmpty();
	}
	
	@Override
	public int size() {
		return this.map.size();
	}
	
	@Override
	@SuppressWarnings("unchecked")
	public V get(Object key) {
		ExpiringValue evalue;
		
		synchronized (this.sync) {
			if (!this.map.containsKey(key))
				return null;
			
			// remove and put to move the entry to the end of the map; helping with finding expired entries
			evalue = this.map.remove(key);
			this.map.put((K)key, new ExpiringValue(evalue.getValue(), System.currentTimeMillis() + this.expirationTimeMillis));
		}
		
		for (MapListener<K, V> listener : this.listeners)
			listener.accessed(new ExpiringHashMapEntry((K)key, evalue.getValue()));
		return evalue.getValue();
	}
	
	@Override
	public V put(K key, V value) {
		ExpiringValue evalue = new ExpiringValue(value, System.currentTimeMillis() + this.expirationTimeMillis);
		ExpiringValue oldValue = this.map.put(key, evalue);
		for (MapListener<K, V> listener : this.listeners)
			listener.added(new ExpiringHashMapEntry(key, value));
		return oldValue == null ? null : oldValue.getValue();
	}
	
	@Override
	public void clear() {
		List<ExpiringHashMapEntry> entries = new LinkedList<>();
		
		synchronized (this.sync) {
			for (Entry<K, ExpiringValue> entry : this.map.entrySet())
				entries.add(new ExpiringHashMapEntry(entry.getKey(), entry.getValue().getValue()));
			this.map.clear();
		}

		for (ExpiringHashMapEntry entry : entries)
			for (MapListener<K, V> listener : this.listeners)
				listener.cleared(entry);
	}
	
	@SuppressWarnings("unchecked")
	@Override
	public V remove(Object key) {
		ExpiringValue evalue;
		
		synchronized (this.sync) {
			if (!this.map.containsKey(key))
				return null;
			evalue = this.map.remove(key);
		}
		
		for (MapListener<K, V> listener : this.listeners)
			listener.removed(new ExpiringHashMapEntry((K)key, evalue.getValue()));
		return evalue.getValue();
	}
	
	@Override
	public void putAll(Map<? extends K, ? extends V> m) {
		throw new UnsupportedOperationException();
	}
	
	@Override
	public int hashCode() {
		return this.map.hashCode();
	}
	
	@Override
	public boolean equals(Object obj) {
		return this.map.equals(obj);
	}
	
	public void exipriationCheck() {
		// since these should be order, we could break out of this loop once we reach an unexpired entry
		List<ExpiringHashMapEntry> entries = new LinkedList<>();
		
		synchronized (this.sync) {
			Iterator<Entry<K, ExpiringValue>> i = this.map.entrySet().iterator();
			for (Entry<K, ExpiringValue> entry = i.next(); i.hasNext(); entry = i.next()) {
				if (entry.getValue().isExpired()) {
					i.remove();
					entries.add(new ExpiringHashMapEntry(entry.getKey(), entry.getValue().getValue()));
				} else {
					// ordered map; ordered by time of entry
					// since expiration timers are constant, we can skip looping
					break;
				}
			}
		}

		for (ExpiringHashMapEntry entry : entries)
			for (MapListener<K, V> listener : this.listeners)
				if (listener instanceof ExpiringMapListener)
					((ExpiringMapListener<K, V>)listener).expired(entry);
	}
	
	public void expire(K key) {
    	if (this.logger.isDebugEnabled())
    		this.logger.debug("expiring key from map: " + key);

		ExpiringValue evalue;
		
    	synchronized (this.sync) {
			if (!this.map.containsKey(key))
				return;
			evalue = this.map.remove(key);
    	}
    	
		for (MapListener<K, V> listener : this.listeners)
			if (listener instanceof ExpiringMapListener)
				((ExpiringMapListener<K, V>)listener).expired(new ExpiringHashMapEntry(key, evalue.getValue()));
	}
	
	@Override
	public String toString() {
		return this.map.toString();
	}
	
	
	
	private class ExpiringHashMapEntry implements Entry<K, V> {
		
		private final K key;
		private V value;
		
		public ExpiringHashMapEntry(K key, V value) {
			this.key = key;
			this.value = value;
		}
		
		@Override
		public K getKey() {
			return this.key;
		}
		
		@Override
		public V getValue() {
			return this.value;
		}
		
		@Override
		public V setValue(V value) {
			return this.value = value;
		}
		
		@Override
		public boolean equals(Object obj) {
			if (obj instanceof Entry<?, ?>) {
				return this.key.equals(((Entry<?, ?>)obj).getKey()) && this.value.equals(((Entry<?, ?>)obj).getValue());
			} else {
				return false;
			}
		}
		
		@Override
		public int hashCode() {
			return this.key.hashCode() + this.value.hashCode();
		}
		
		@Override
		public String toString() {
			return "{" + this.key + ", " + this.value + "}";
		}
		
	}
	
	private class ExpiringValue {
	
		private long expirationTimeInMillis;
		private V value;
		
		public ExpiringValue(V value, long expirationTimeInMillis) {
			this.value = value;
			this.expirationTimeInMillis = expirationTimeInMillis;
		}
		
		public V getValue() {
			return this.value;
		}
		
		public boolean isExpired() {
			return this.expirationTimeInMillis <= System.currentTimeMillis();
		}
		
		@Override
		public String toString() {
			return this.value.toString();
		}
		
	}

}
